#include <stdio.h>

extern "C" {
  #include <stdlib.h>
  #include <string.h>
  #include <inttypes.h>
  #include "twi2.h"
}


#include <DS3231.h>
#include "twi2.h"
#include <Wire2.h>
#include "Time.h"

DS3231 Clock;



byte clock_flag;


time_t a1time(void)
{
 tmElements_t tx; 
 byte year;
 byte month;
 byte date;
 byte hour;
 byte minute;
 byte second;
 byte dow;
 byte abits;
 bool aday;
 bool ah12;
 bool apm;
  
 Clock.getTime(year,month,date,dow,hour,minute,second);  
 Clock.getA1Time(date,hour,minute,second,abits,aday,ah12,apm);

 tx.Second=second;
 tx.Minute=minute;
 tx.Hour=hour;
 tx.Day=date;
 tx.Month=month;
 tx.Year=CalendarYrToTm(year+2000);
 tx.Wday=0;


  Serial.print("set alarm Day ");
  Serial.print(tx.Day);
  
  Serial.print(" Hour ");
  Serial.print(tx.Hour);
  
  Serial.print(" Min ");
  Serial.print(tx.Minute);

  Serial.print(" Sec ");
  Serial.println(tx.Second);


 return(makeTime(tx)); 
}  


time_t nowtime(void)
{
 tmElements_t tx; 
 byte year;
 byte month;
 byte date;
 byte dow;
 byte hour;
 byte minute;
 byte second;
  
 Clock.getTime(year,month,date,dow,hour,minute,second);

 tx.Second=second;
 tx.Minute=minute;
 tx.Hour=hour;
 tx.Day=date;
 tx.Month=month;
 tx.Year=CalendarYrToTm(year+2000);
 tx.Wday=dow;

 return(makeTime(tx)); 
}  


unsigned long targetcount;

void clock_next()
{
 time_t now; 
 signed long diff;
 time_t target;
 unsigned long count;
 tmElements_t tx; 
 
  Serial.print("free memory ");
Serial.print(get_free_memory()); 
Serial.println();

 
 now=nowtime();
 diff=now-bootblock.basetime;
 count=diff/bootblock.period;
 
 Serial.print("count ");
 Serial.println(count);
 
 while(1)
 {
  count++;
  targetcount=count;
  target=count*bootblock.period+bootblock.basetime;
  breakTime(target,tx);
  Clock.checkIfAlarm(1);
  Clock.setA1Time(tx.Day,tx.Hour,tx.Minute,tx.Second,0,0,0,0);
  
  Serial.print("set alarm count ");
  Serial.print(count);
  
  Serial.print("Day ");
  Serial.print(tx.Day);
  
  Serial.print(" Hour ");
  Serial.print(tx.Hour);
  
  Serial.print(" Min ");
  Serial.print(tx.Minute);

  Serial.print(" Sec ");
  Serial.println(tx.Second);
  
  now=nowtime();
  diff=now-target;
  if(diff<0) break;
 }
 
 Serial.print("a1time  ");
 Serial.println(a1time());
 
 Serial.print("nowtime  ");
 Serial.println(nowtime());
 
}  



void powerup_clock(void)
{
 pinMode(pin_clock_power, OUTPUT);  
 digitalWrite(pin_clock_power, HIGH);
// Wire.begin();  
  
}  


void powerdown_clock(void)
{
// twi_stop();
//TWCR = bit(TWEN) | bit(TWIE) | bit(TWEA) | bit(TWINT);
 digitalWrite(pin_clock_power, LOW);
 pinMode(pin_clock_power, INPUT);   
  
}  


void clock_setup() 
{
 Wire.begin();

 pinMode(3, INPUT_PULLUP);
 Clock.turnOffAlarm(2);
 Clock.turnOnAlarm(1);
 Clock.checkIfAlarm(1);

 clock_next();

 attachInterrupt(1, clock_interrupt, FALLING);
 clock_flag=0;
}


void clock_reset()
{
 clock_next();  
  
} 




void clock_interrupt(void)
{
 clock_flag=1; 
}  

void clock_clear(void)
{
 clock_flag=0; 
 Serial.print("clock clear ");
 Serial.println(digitalRead(3));
 Serial.flush();
}  


extern unsigned int __bss_end;
extern void *__brkval;

int get_free_memory()
{
  int free_memory;

  if((int)__brkval == 0)
    free_memory = ((int)&free_memory) - ((int)&__bss_end);
  else
    free_memory = ((int)&free_memory) - ((int)__brkval);

  return free_memory;
}


// is it going to be more than secs before the alarm goes off

int clock_sleep_ok(int secs)
{
 time_t al;
 time_t diff;
 time_t now;

 al=a1time();
 now=nowtime();
 diff=al-now;
 return(diff>secs);
}  


void clock_interruptcheck(void)
{
 char buffer[12];
 int  size;
 time_t now;
 time_t diff;
 unsigned long count;
 unsigned int  location;
 unsigned long wrap;
 int countmod;
 int lastmod;
 int i;
 
 
 if(clock_flag)
 {
  clock_flag=0; // put this first in case interrupt goes off during RXdata()

  RXdata();
 
  now=a1time();
  diff=now-bootblock.basetime;
  count=diff/bootblock.period;

  bootblock.targetmiss|=(count!=targetcount);

if(count!=targetcount)
{
 Serial.print("targetmiss ");
 Serial.print(count);
 Serial.print(" ");
 Serial.print(targetcount);
 
  
}

  memset(buffer,0xDE,sizeof(buffer));
  wrap=bootblock.memorysize/bootblock.recordsize;
  countmod=count % wrap;
  if((count-bootblock.lastgoodvalue)<wrap)
  {
   lastmod=(bootblock.lastgoodvalue+1) % wrap;
  } 
  else
  {
   lastmod=(count+1) % wrap;
  }  
  i=lastmod;


  if(lastmod!=countmod)
  {
   Serial.print("blanking ");
   Serial.print(lastmod);
   Serial.print(" to ");
   Serial.println(countmod);
  
   bootblock.blank1=lastmod;
   bootblock.blank2=countmod;
   while(1)
   {
    if(i==countmod) break;
    if(i>=wrap) i=0;
    location=i*bootblock.recordsize;
    i2c_eeprom_write_page(0x57,location,(byte*)buffer,bootblock.recordsize);
    i++;
   }  
  }
  
  size=packsensors(buffer,sizeof(buffer));  
  location=(countmod*size);

//memcpy(buffer,&count,sizeof(count));

  i2c_eeprom_write_page(0x57,location,(byte*)buffer,size);

  bootblock.lastgoodvalue=count;
  writelastgoodvalue();

  Serial.print("write data count ");
  Serial.println(count);


  clock_next();
 }
}


void ReadDS3231()
{
  byte year;
  byte month;
  byte date;
  byte dow;
  byte hour;
  byte minute;
  byte second;
  float temperature;
  bool  h12;
  bool PM;
  
  Clock.getTime(year,month,date,dow,hour,minute,second);  
  temperature=Clock.getTemperature();
  
  Serial.print("20");
  Serial.print(year,DEC);
  Serial.print('-');
  Serial.print(month,DEC);
  Serial.print('-');
  Serial.print(date,DEC);
  Serial.print(' ');
  Serial.print(hour,DEC);
  Serial.print(':');
  Serial.print(minute,DEC);
  Serial.print(':');
  Serial.print(second,DEC);
  Serial.println();
  Serial.print("Temperature=");
  Serial.println(temperature); 
}


void clock_loop() 
{
// ReadDS3231();
// Serial.print("input 2 ");
// Serial.print(digitalRead(3));
// Serial.print(Clock.checkIfAlarm(1));
// Serial.println();
  clock_interruptcheck();
 // delay(1000);
}



void cmdwriteseconds(void)
{
 char * arg;
 arg = sCmd.next();
 if(arg!=NULL)
 {
  Clock.setSecond(atoi(arg));
 }
}



void cmdwriteminutes(void)
{
 char * arg;
 arg = sCmd.next();
 if(arg!=NULL)
 {
  Clock.setMinute(atoi(arg));
 }
}




void cmdwritehours(void)
{
 char * arg;
 arg = sCmd.next();
 if(arg!=NULL)
 {
  Clock.setHour(atoi(arg));
 }
}

void cmdwritedate(void)
{
 char * arg;
 arg = sCmd.next();
 if(arg!=NULL)
 {
  Clock.setDate(atoi(arg));
 }
}



void cmdwritemonth(void)
{
 char * arg;
 arg = sCmd.next();
 if(arg!=NULL)
 {
  Clock.setMonth(atoi(arg));
 }
}


void cmdwriteyear(void)
{
 char * arg;
 arg = sCmd.next();
 if(arg!=NULL)
 {
  Clock.setYear(atoi(arg));
 }
}




