// David Pilling 6th June 2014
// Receive Oregon v2 packets and print out values
// Also receives packets from my own liquid depth sensor

// the conversion from raw packet to numeric values was inspired by
// http://www.disk91.com/2013/technology/hardware/oregon-scientific-sensors-with-raspberry-pi/
// I suspect none of that code remains though


// I made two key changes to the code below
// 1) make preamble 24 bits (as others have done)
// 2) packet length 159 bits - one of my sensors does not seem to send 160 bits

// Oregon V2 decoder added - Dominique Pierre
// Oregon V3 decoder revisited - Dominique Pierre
// New code to decode OOK signals from weather sensors, etc.
// 2010-04-11 <jcw@equi4.com> http://opensource.org/licenses/mit-license.php
// $Id: ookDecoder.pde 5331 2010-04-17 10:45:17Z jcw $

class DecodeOOK {
protected:
    byte total_bits, bits, flip, state, pos, data[25];

    virtual char decode (word width) =0;
    
public:

    enum { UNKNOWN, T0, T1, T2, T3, OK, DONE };

    DecodeOOK () { resetDecoder(); }

    bool nextPulse (word width) {
        if (state != DONE)
        
            switch (decode(width)) {
                case -1: resetDecoder(); break;
                case 1:  done(); break;
            }
        return isDone();
    }
    
    bool isDone () const { return state == DONE; }

    const byte* getData (byte& count) const {
        count = pos;
        return data; 
    }
    
    void resetDecoder () {
        total_bits = bits = pos = flip = 0;
        state = UNKNOWN;
    }
    
    // add one bit to the packet data buffer
    
    virtual void gotBit (char value) {
        total_bits++;
        byte *ptr = data + pos;
        *ptr = (*ptr >> 1) | (value << 7);

        if (++bits >= 8) {
            bits = 0;
            if (++pos >= sizeof data) {
                resetDecoder();
                return;
            }
        }
        state = OK;
    }
    
    // store a bit using Manchester encoding
    void manchester (char value) {
        flip ^= value; // manchester code, long pulse flips the bit
        gotBit(flip);
    }
    
    // move bits to the front so that all the bits are aligned to the end
    void alignTail (byte max =0) {
        // align bits
        if (bits != 0) {
            data[pos] >>= 8 - bits;
            for (byte i = 0; i < pos; ++i)
                data[i] = (data[i] >> bits) | (data[i+1] << (8 - bits));
            bits = 0;
        }
        // optionally shift bytes down if there are too many of 'em
        if (max > 0 && pos > max) {
            byte n = pos - max;
            pos = max;
            for (byte i = 0; i < pos; ++i)
                data[i] = data[i+n];
        }
    }
    
    void reverseBits () {
        for (byte i = 0; i < pos; ++i) {
            byte b = data[i];
            for (byte j = 0; j < 8; ++j) {
                data[i] = (data[i] << 1) | (b & 1);
                b >>= 1;
            }
        }
    }
    
    void reverseNibbles () {
        for (byte i = 0; i < pos; ++i)
            data[i] = (data[i] << 4) | (data[i] >> 4);
    }
    
    void done () {
        while (bits)
            gotBit(0); // padding
        state = DONE;
    }
};

// 433 MHz decoders


class OregonDecoderV2 : public DecodeOOK {
public:
    OregonDecoderV2() {}
    
    // add one bit to the packet data buffer
    virtual void gotBit (char value) {
        if(!(total_bits & 0x01))
        {
            data[pos] = (data[pos] >> 1) | (value ? 0x80 : 00);
        }
        total_bits++;
        pos = total_bits >> 4;
        if (pos >= sizeof data) {
            resetDecoder();
            return;
        }
        state = OK;
    }
    
    virtual char decode (word width) {
        if (200 <= width && width < 1200) {
            byte w = width >= 700;
            switch (state) {
                case UNKNOWN:
                    if (w != 0) {
                        // Long pulse
                        ++flip;
                    } else if (24 <= flip) {
                        // Short pulse, start bit
                        flip = 0;
                        state = T0;
                    } else {
                      // Reset decoder
                        return -1;
                    }
                    break;
                case OK:
                    if (w == 0) {
                        // Short pulse
                        state = T0;
                    } else {
                        // Long pulse
                        manchester(1);
                    }
                    break;
                case T0:
                    if (w == 0) {
                      // Second short pulse
                        manchester(0);
                    } else {
                        // Reset decoder
                        return -1;
                    }
                    break;
            }
        } else {
            return -1;
        }
        return total_bits == 159 ? 1: 0;
    }
};


OregonDecoderV2 orscV2;



volatile word pulse;


void ext_int_1(void) {
    static word last;
    // determine the pulse length in microseconds, for either polarity
    pulse = micros() - last;
    last += pulse;
}


char itox(int n)
{
 const char array[]={'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
 return(array[n]); 
}  




void reportSerial (const char* s, class DecodeOOK& decoder)
{
 byte pos;
 char string[64];
 char * p=string;

 const byte* data = decoder.getData(pos);

 strcpy(string,s);
 p+=strlen(s);

 *p++=' ';
 
 for(byte i = 0; i < pos; ++i) 
 {
  *p++=itox(data[i] >> 4);
  *p++=itox(data[i] & 0x0F);
 }
    
 *p=0;
 
 Serial.println(string);
    
 decoder.resetDecoder();
 
 char * pt=string+5;
 char temp[8];
 char humid[4];
 
 temp[0] = pt[10] ; temp[1] = pt[11] ; temp[2] = pt[8] ; temp[3] = '\0';
 humid[0] = pt[15] ; humid[1] = pt[12]; humid[2] = '0' ; humid[3] = '\0';
// nybble 4 = (1<<id)

 if(strncmp(pt,"1A2D",4)==0)
 {
  Serial.print("temp ");
  Serial.println(temp);
 
  Serial.print("humid ");
  Serial.println(humid);
 }
 else
 if(strncmp(pt,"EA4C",4)==0) 
 {
  Serial.print("temp ");
  Serial.println(temp);    
 } 
 else
 if(strncmp(pt,"1ABC",4)==0) 
 {
  // 1ABC2053512300002D00
 
  strncpy(temp,pt+6,2);
  temp[2]=0; 
  Serial.print("vcc ");
  Serial.print(temp);
  
  strncpy(temp,pt+8,2); 
  temp[2]=0; 
  Serial.print(" humid ");
  Serial.print(temp);
  
  strncpy(temp,pt+10,2); 
  temp[2]=0; 
  Serial.print(" temp ");
  Serial.print(temp);
  
  strncpy(temp,pt+12,4);
  temp[4]=0; 
  Serial.print(" freq ");
  Serial.println(temp);

 }
}






int sensorsread;
int temperatures[MAX_SENSORS];
int humidities[MAX_SENSORS];
int frequencies[MAX_SENSORS];
int vccs[MAX_SENSORS];


int packsensors(char * buffer,int bsize)
{
 int    i;
 char * p=buffer;

 memset(buffer,0xAB,bsize);
 
 for(i=0;i<bootblock.nsensors;i++)
 {
  if(bootblock.sensormodel[i]==0xEA4C)
  {
   *p++=temperatures[i]/10; 
  }
  else
  if(bootblock.sensormodel[i]==0x1A2D)
  {
   *p++=temperatures[i]/10;     
   *p++=humidities[i];     
  }
  else
  if(bootblock.sensormodel[i]==0x1ABC)
  {
   *p++=temperatures[i];     
   *p++=humidities[i];     
   *p++=vccs[i];     
   *p++=frequencies[i]/10;     
  }  
 } 
 
 return(p-buffer);
}



int findsensor(int model,int channel)
{
 int i;

 for(i=0;i<bootblock.nsensors;i++)
 {
  if(bootblock.sensormodel[i]==model && bootblock.sensorchannel[i]==channel) return(i);
 }
 
 return(-1);
}  



void processSerial (class DecodeOOK& decoder)
{
 byte pos;
 int  sensor;
 const byte* data = decoder.getData(pos);

 
 decoder.resetDecoder();
 

 int temp;
 int humid;
 int id;
 
 
 temp = (data[5]>>4)*100+ (data[5] & 0xF)*10+ (data[4]>>4);
 if(data[6] & 0xF) temp=-temp;
 humid = (data[7] & 0xF)*10+ (data[6] >>4);
 id=(data[0]<<8)+data[1];

/*
 int i;
 for(i=0;i<8;i++)
 {
  Serial.print(i);
  Serial.print(" ");
  Serial.print(data[i],HEX);
  Serial.println();
 }  

Serial.print(" id ");
Serial.print(id,HEX);
*/
 
 sensor=findsensor(id,data[2]>>4);
 
Serial.print(" sensor ");
Serial.print(sensor);
Serial.println();

 if(sensor>=0)
 {
  if(vccs[sensor]==-1)
  {
   vccs[sensor]=0;
   sensorsread++;
  }
  
  if(id==0x1A2D)
  {
    // temp humid
    Serial.print(" temp ");
   Serial.print(temp);

    Serial.print(" humid ");
   Serial.print(humid);
    humidities[sensor]=humid;
    temperatures[sensor]=temp;
  }
  else
  if(id==0xEA4C) 
  {
   // temp
    temperatures[sensor]=temp;   
   Serial.print(" temp ");
   Serial.print(temp);

  } 
  else
  if(id==0x1ABC) 
  {
  // 1ABC20 53512300002D00
  
   vccs[sensor]=(data[3]>>4)*10+(data[3] & 0xF);
   humidities[sensor]=(data[4]>>4)*10+(data[4] & 0xF);
   temperatures[sensor]=(data[5]>>4)*10+(data[5] & 0xF);
   frequencies[sensor]=(data[6]>>4)*1000+(data[6] & 0xF)*100+(data[7]>>4)*10+(data[7] & 0xF);
  
  
  }
   Serial.println();
     Serial.print(" temp ");
     Serial.print(temperatures[sensor]);
     Serial.print(" humid ");
     Serial.print(humidities[sensor]);
     Serial.print(" vcc ");
     Serial.print(vccs[sensor]);
     Serial.print(" freq ");
     Serial.print(frequencies[sensor]);
     
 }
 
 Serial.println();
}









void RXdata()
{
 int i; 
 unsigned long time=millis()+((unsigned long)55)*((unsigned long)1000);

 sensorsread=0;
 
 Serial.println("RX data start");
 
 for(i=0;i<bootblock.nsensors;i++) vccs[i]=-1;
 
 while((sensorsread<bootblock.nsensors) && (millis()<time))
 {
  cli();
  word p = pulse;
  pulse = 0;
  sei();

   
  if (p != 0) 
  {
   if(orscV2.nextPulse(p))
            processSerial(orscV2);  
           
  }  
 } 
 
  Serial.print(" sensors read ");
  Serial.print(sensorsread);
  Serial.print(" ");
  Serial.print((time-millis())/((unsigned long)1000));
  Serial.println(" RX data end");
}  








void RXsetup () 
{
 attachInterrupt(0, ext_int_1, CHANGE);
}



void RXloop () 
{
/*
    cli();
    word p = pulse;
    
    pulse = 0;
    sei();

   
    if (p != 0) 
    {
        if (orscV2.nextPulse(p))
            reportSerial("OSV2", orscV2);  
           
    }
  */  
}

